module Main where

import Test.Hspec (hspec, it, shouldBe)
import Data.Function ((&))
import Data.Char (digitToInt)
import Data.Bits (shift)

parse :: String -> [Int]
parse = map digitToInt

bitListToInt :: [Int] -> Int
bitListToInt [] = 0
bitListToInt (b:bs) = b `shift` length bs + bitListToInt bs

boolListToInt :: [Bool] -> Int
boolListToInt = bitListToInt . map fromEnum

gammaEpsilon :: [[Int]] -> (Int, Int)
gammaEpsilon xs = (boolListToInt mostCommon, boolListToInt $ map not mostCommon)
  where
    bitsUsed = foldl1 (zipWith (+)) xs
    mostCommon = map (> (length xs `div` 2)) bitsUsed

powerConsumption :: [[Int]] -> Int
powerConsumption = uncurry (*) . gammaEpsilon

rating :: (Int -> Int -> Bool) -> [[Int]] -> Int
rating compare = let
    rating' _ [x] = x
    rating' n xs = rating' (n+1) filtered
      where
        bitSum = sum (map (!!n) xs)
        mostCommon = fromEnum $ bitSum `compare` (length xs - bitSum)
        filtered = filter ((==mostCommon) . (!!n)) xs
  in bitListToInt . rating' 0

oxyRating = rating (>=)
co2Rating = rating (<)

runTests = do
  hspec $ do
    it "parses binary numbers" $ do
      parse "00100" `shouldBe` [0,0,1,0,0]
      parse "10011" `shouldBe` [1,0,0,1,1]
    it "converts bits to integers" $ do
      boolListToInt [False] `shouldBe` 0
      boolListToInt [True] `shouldBe` 1
      boolListToInt [False,True] `shouldBe` 1
      boolListToInt [True,True] `shouldBe` 3
      boolListToInt [True,True,False,True,False] `shouldBe` 26
    it "calculates γ and ε" $ do
      gammaEpsilon testDiagnostic `shouldBe` (22, 9)
    it "calculates power consumption" $ do
      powerConsumption testDiagnostic `shouldBe` 198
    it "calculates oxygen generator rating" $ do
      oxyRating testDiagnostic `shouldBe` 23
    it "calculates CO₂ scrubber rating" $ do
      co2Rating testDiagnostic `shouldBe` 10
  where
    testInput = ["00100", "11110", "10110", "10111", "10101", "01111", "00111", "11100", "10000", "11001", "00010", "01010"]
    testDiagnostic = map parse testInput

main :: IO ()
main = do
  runTests

  input <- getContents & fmap lines
  let diagnostic = map parse input

  putStr "--\nPart 1: "
  print $ powerConsumption diagnostic
  putStr "Part 2: "
  print $ oxyRating diagnostic * co2Rating diagnostic
