{-# LANGUAGE TupleSections #-}
module Main where

import Test.Hspec(hspec, describe, it, shouldBe)
import Data.Maybe
import Data.Either
import Data.Array(Array,(!))
import qualified Data.Array as A
import Data.Void
import Text.Megaparsec hiding (State)
import Text.Megaparsec.Char

type Parser = Parsec Void String

type Call = [Int]
data Square = Unmarked Int | Marked Int
  deriving (Eq, Show)
type Board = Array (Int, Int) Square

mkBoard :: [[Int]] -> Board
mkBoard nums = A.listArray ((0,0), (4,4)) $ map Unmarked $ concat nums

squareVal :: Square -> Int
squareVal (Unmarked n) = n
squareVal (Marked n) = n

marked :: Square -> Bool
marked (Unmarked _) = False
marked (Marked _) = True

unmarked :: Square -> Bool
unmarked = not . marked

markBoard :: Int -> Board -> Board
markBoard n b = mark <$> b
  where
    mark sq
      | squareVal sq == n = Marked n
      | otherwise         = sq

checkLine :: Board -> Maybe Int
checkLine b = if check getRow || check getCol then Just $ scoreBoard b else Nothing
  where
    check getter = any getter [0..4]
    getRow i = all (marked . (b!) . (i,)) [0..4]
    getCol j = all (marked . (b!) . (,j)) [0..4]
    scoreBoard = sum . map squareVal . filter unmarked . A.elems

runGame :: Call -> [Board] -> [Maybe Int]
runGame [] _ = []
runGame (x:xs) bs = result : runGame xs bsNext
  where
    bsMarked = map (markBoard x) bs
    bsScored = map (\b -> (checkLine b, b)) bsMarked
    result = listToMaybe $ mapMaybe (fmap (x*) . fst ) bsScored
    bsNext = map snd $ filter (isNothing . fst) bsScored

results :: Call -> [Board] -> [Int]
results cs bs = catMaybes $ runGame cs bs

gameP :: Parser (Call, [Board])
gameP = (,) <$> call <* count 2 newline <*> boards
  where
    integer = read <$> some digitChar
    call = integer `sepBy` char ','
    row = hspace *> integer `sepBy` hspace1
    board = count 5 $ row <* newline
    boards = map mkBoard <$> board `sepEndBy` newline

------------------------------------------------------------------------------

main = do
  runTests

  Right (call, boards) <- parse gameP "(stdin)" <$> getContents
  let gameResults = results call boards

  putStr "--\nPart 1: "
  print $ head gameResults
  putStr "Part 2: "
  print $ last gameResults

------------------------------------------------------------------------------

runTests = hspec $ do
  describe "parser" $ do
    it "parses a game" $ do
      parse gameP "" testInput
        `shouldBe` Right ([7,4,9,5,11,17,23,2,0,14,21,24,10,16,13,6,15,25,12,22,18,20,8,19,3,26,1], map mkBoard [[[22,13,17,11,0],[8,2,23,4,24],[21,9,14,16,7],[6,10,3,18,5],[1,12,20,15,19]],[[3,15,0,2,22],[9,18,13,17,5],[19,8,7,25,23],[20,11,10,24,4],[14,21,16,12,6]],[[14,21,17,24,4],[10,16,15,9,19],[18,8,23,26,20],[22,11,13,6,5],[2,0,12,3,7]]])
  describe "game" $ do
    it "marks a board" $ do
      markBoard 2 (A.listArray ((0,0),(1,1)) $ map Unmarked [1..4])
        `shouldBe` A.listArray ((0,0),(1,1)) [Unmarked 1, Marked 2, Unmarked 3, Unmarked 4]
      markBoard 3 (A.listArray ((0,0),(1,1)) [Unmarked 1, Marked 2, Unmarked 3, Unmarked 4])
        `shouldBe` A.listArray ((0,0),(1,1)) [Unmarked 1, Marked 2, Marked 3, Unmarked 4]
    it "checks a line" $ do
      checkLine (A.listArray ((0,0),(4,4)) $ map Unmarked [1..25])
        `shouldBe` Nothing
      checkLine (A.listArray ((0,0),(4,4)) ([Marked 1, Marked 2, Marked 3, Marked 4, Marked 5] ++ map Unmarked [6..25]))
        `shouldBe` Just 310
      checkLine (A.listArray ((0,0),(4,4)) ([if n `mod` 5 == 0 then Marked n else Unmarked n | n <- [1..25]]))
        `shouldBe` Just 250
    it "runs a game" $ do
      head (uncurry results testGame) `shouldBe` 4512
      last (uncurry results testGame) `shouldBe` 1924

testInput = unlines [
        "7,4,9,5,11,17,23,2,0,14,21,24,10,16,13,6,15,25,12,22,18,20,8,19,3,26,1",
        "",
        "22 13 17 11  0",
        " 8  2 23  4 24",
        "21  9 14 16  7",
        " 6 10  3 18  5",
        " 1 12 20 15 19",
        "",
        "3  15  0  2 22",
        "9  18 13 17  5",
        "19  8  7 25 23",
        "20 11 10 24  4",
        "14 21 16 12  6",
        "",
        "14 21 17 24  4",
        "10 16 15  9 19",
        "18  8 23 26 20",
        "22 11 13  6  5",
        "2  0  12  3  7"]
testGame = fromRight ([], []) $ parse gameP "" testInput
