module Main where

import Test.Hspec
import Data.Either
import Data.Void
import Data.Map.Strict as M
import Text.Megaparsec
import Text.Megaparsec.Char


type Parser = Parsec Void String

type Point = (Int, Int)
type Vent = ((Int, Int), (Int, Int))
type Seafloor = M.Map Point Int


linesP :: Parser [Vent]
linesP = line `sepEndBy` newline
  where
    integer :: Parser Int
    integer = read <$> some digitChar
    pair = (,) <$> integer <* char ',' <*> integer
    line = (,) <$> pair <* string " -> " <*> pair


plotVentHV :: Vent -> [Point]
plotVentHV ((a, b), (c, d)) = zip xs ys
  where
    (xs, ys)
      | a == c = (repeat a, [min b d..max b d])
      | b == d = ([min a c..max a c], repeat b)
      | otherwise = ([], [])

plotVent :: Vent -> [Point]
plotVent ((a, b), (c, d)) = zip xs ys
  where
    xs = series a c
    ys = series b d
    series i j = case compare i j of
      EQ -> repeat i
      LT -> [i..j]
      GT -> [i,i-1..j]

mapVent :: (Vent -> [Point]) -> Seafloor -> Vent -> Seafloor
mapVent plot fl l = M.unionWith (+) fl updates
  where updates = M.fromList [(k, 1) | k <- plot l]

dangerLevel :: (Vent -> [Point]) -> [Vent] -> Int
dangerLevel plot ls = M.size $ M.filter (>=2) fl
  where
    fl = Prelude.foldl (mapVent plot) M.empty ls

------------------------------------------------------------------------------

main = do
  runTests

  input <- getContents
  let vents = fromRight [] $ parse linesP "(stdin)" input

  putStr "--\nPart 1: "
  print $ dangerLevel plotVentHV vents
  putStr "Part 2: "
  print $ dangerLevel plotVent vents

------------------------------------------------------------------------------

runTests = hspec $ do
  describe "parser" $ do
    it "parses without error" $ do
      parse linesP "(tests)" testInput `shouldSatisfy` isRight
    it "returns the correct result" $ do
      parse linesP "(tests)" testInput `shouldBe` Right testSample
  describe "solution" $ do
    it "plots h/v vents" $ do
      plotVentHV ((0, 0), (0, 3)) `shouldBe` [(0,0),(0,1),(0,2),(0,3)]
      plotVentHV ((0, 3), (0, 0)) `shouldBe` [(0,0),(0,1),(0,2),(0,3)]
      plotVentHV ((3, 0), (0, 0)) `shouldBe` [(0,0),(1,0),(2,0),(3,0)]
    it "plots all vents" $ do
      plotVent ((0, 0), (0, 3)) `shouldBe` [(0,0),(0,1),(0,2),(0,3)]
      plotVent ((0, 3), (0, 0)) `shouldBe` [(0,3),(0,2),(0,1),(0,0)]
      plotVent ((3, 0), (0, 0)) `shouldBe` [(3,0),(2,0),(1,0),(0,0)]
      plotVent ((0, 0), (3, 3)) `shouldBe` [(0,0),(1,1),(2,2),(3,3)]
      plotVent ((0, 3), (3, 0)) `shouldBe` [(0,3),(1,2),(2,1),(3,0)]
    it "marks vents" $ do
      mapVent plotVentHV M.empty ((0, 0), (0, 1)) `shouldBe` M.fromList [(k, 1) | k <- [(0,0),(0,1)]]
      mapVent plotVentHV (M.fromList [((0,0), 1), ((1,1), 1)]) ((0, 0), (0, 1))
        `shouldBe` M.fromList [((0,0), 2), ((0,1), 1), ((1,1), 1)]
      mapVent plotVent M.empty ((0, 0), (1, 1)) `shouldBe` M.fromList [(k, 1) | k <- [(0,0),(1,1)]]
    it "understands danger" $ do
      dangerLevel plotVentHV testSample `shouldBe` 5
      dangerLevel plotVent testSample `shouldBe` 12

testInput = "0,9 -> 5,9\n\
            \8,0 -> 0,8\n\
            \9,4 -> 3,4\n\
            \2,2 -> 2,1\n\
            \7,0 -> 7,4\n\
            \6,4 -> 2,0\n\
            \0,9 -> 2,9\n\
            \3,4 -> 1,4\n\
            \0,0 -> 8,8\n\
            \5,5 -> 8,2\n"

testSample
  = [((0, 9), (5, 9)), ((8, 0), (0, 8)), ((9, 4), (3, 4)), ((2, 2), (2, 1)),
     ((7, 0), (7, 4)), ((6, 4), (2, 0)), ((0, 9), (2, 9)), ((3, 4), (1, 4)),
     ((0, 0), (8, 8)), ((5, 5), (8, 2))]
