module Main where

import Test.Hspec
import qualified Data.List as L
import Data.Map.Strict(Map)
import qualified Data.Map.Strict as M
import Data.Set(Set,(\\))
import qualified Data.Set as S
import Data.Maybe
import Data.Either
import Data.Void
import Text.Megaparsec
import Text.Megaparsec.Char

type Parser = Parsec Void String

type Signal = Set Char
type Note = ([Signal], [Signal])
type SignalMap = Map Signal Int

notesP :: Parser [Note]
notesP = note `sepEndBy` newline
  where
    note = (,) <$> manyWords <* string "| " <*> manyWords
    manyWords = word `sepEndBy` hspace1
    word = S.fromList <$> some lowerChar

countKnown :: [Note] -> Int
countKnown notes = length $ filter ((`elem` [2,4,3,7]) . length) $ concatMap snd notes

inferDigits :: [Signal] -> SignalMap
inferDigits signals = M.fromList [(sig0, 0), (sig1, 1), (sig2, 2), (sig3, 3), (sig4, 4), (sig5, 5), (sig6, 6), (sig7, 7), (sig8, 8), (sig9, 9)]
  where
    byLength n = L.find ((==n) . length)
    Just sig1 = byLength 2 signals
    Just sig4 = byLength 4 signals
    Just sig7 = byLength 3 signals
    Just sig8 = byLength 7 signals

    -- 6-segment digits: 0, 6, 9
    with6 = filter ((==6) . length) signals
    almost9 = S.union sig4 sig7
    -- only digit with one more segment than 4∪7 is 9
    ([sig9], with6') = L.partition ((==1) . length . (\\almost9)) with6
    -- 1 shares two segments with 0, only one with 6
    ([sig0], [sig6]) = L.partition ((==2) . length . S.intersection sig1) with6'

    -- 5-segment digits: 2, 3, 5
    with5 = filter ((==5) . length) signals
    -- of five-segment digits, only 2 shares two segments with 4
    ([sig2], with5') = L.partition ((==2) . length . S.intersection sig4) with5
    -- of the rest, only 3 shares two segments with 1
    ([sig3], [sig5]) = L.partition ((==2) . length . S.intersection sig1) with5'

toDecimal :: [Int] -> Int
toDecimal = toDecimal' . reverse
  where
    toDecimal' [] = 0
    toDecimal' (n:ns) = n + 10 * toDecimal' ns

decode :: Note -> Int
decode (signals, outputs) = toDecimal $ map (fromJust . flip M.lookup digitMap) outputs
  where digitMap = inferDigits signals

------------------------------------------------------------------------------

main = do
  runTests

  Right notes <- parse notesP "(stdin)" <$> getContents

  putStr "--\nPart 1: "
  print $ countKnown notes
  putStr "Part 2: "
  print $ sum $ map decode notes

------------------------------------------------------------------------------

runTests = hspec $ do
  it "counts known patterns" $ do
    countKnown testSample `shouldBe` 26
  it "converts list to decimal" $ do
    toDecimal [5,3,5,3] `shouldBe` 5353
    toDecimal [1,2,3,4] `shouldBe` 1234
  it "infers digits" $ do
    inferDigits (S.fromList <$> ["acedgfb", "cdfbe", "gcdfa", "fbcad", "dab", "cefabd", "cdfgeb", "eafb", "cagedb", "ab"])
      `shouldBe` M.fromList [(S.fromList "acedgfb", 8),
                             (S.fromList "cdfbe", 5),
                             (S.fromList "gcdfa", 2),
                             (S.fromList "fbcad", 3),
                             (S.fromList "dab", 7),
                             (S.fromList "cefabd", 9),
                             (S.fromList "cdfgeb", 6),
                             (S.fromList "eafb", 4),
                             (S.fromList "cagedb", 0),
                             (S.fromList "ab", 1)]
  it "decodes notes" $ do
    decode (S.fromList <$> ["acedgfb", "cdfbe", "gcdfa", "fbcad", "dab", "cefabd", "cdfgeb", "eafb", "cagedb", "ab"],
            S.fromList <$> ["cdfeb", "fcadb", "cdfeb", "cdbaf"])
      `shouldBe` 5353
testInput = unlines [
        "be cfbegad cbdgef fgaecd cgeb fdcge agebfd fecdb fabcd edb | fdgacbe cefdb cefbgd gcbe",
        "edbfga begcd cbg gc gcadebf fbgde acbgfd abcde gfcbed gfec | fcgedb cgb dgebacf gc",
        "fgaebd cg bdaec gdafb agbcfd gdcbef bgcad gfac gcb cdgabef | cg cg fdcagb cbg",
        "fbegcd cbd adcefb dageb afcb bc aefdc ecdab fgdeca fcdbega | efabcd cedba gadfec cb",
        "aecbfdg fbg gf bafeg dbefa fcge gcbea fcaegb dgceab fcbdga | gecf egdcabf bgf bfgea",
        "fgeab ca afcebg bdacfeg cfaedg gcfdb baec bfadeg bafgc acf | gebdcfa ecba ca fadegcb",
        "dbcfg fgd bdegcaf fgec aegbdf ecdfab fbedc dacgb gdcebf gf | cefg dcbef fcge gbcadfe",
        "bdfegc cbegaf gecbf dfcage bdacg ed bedf ced adcbefg gebcd | ed bcgafe cdgba cbgef",
        "egadfb cdbfeg cegd fecab cgb gbdefca cg fgcdab egfdb bfceg | gbdfcae bgc cg cgb",
        "gcafb gcf dcaebfg ecagb gf abcdeg gaef cafbge fdbac fegbdc | fgae cfgab fg bagce"
        ]
testSample = fromRight [] $ parse notesP "(test)" testInput
