module Main where

import Test.Hspec
import Data.Array(Array,bounds,indices,inRange,(!))
import qualified Data.Array as A
import Data.Set(Set)
import qualified Data.Set as S
import Data.Char(digitToInt)
import Data.List(sort)
import Data.Void(Void)
import Text.Megaparsec
import Text.Megaparsec.Char

type Parser = Parsec Void String

type Point = (Int, Int)
type Heightmap = Array Point Int
type PointSet = Set Point

mkHeightmap :: [[Int]] -> Heightmap
mkHeightmap xs = A.listArray ((1,1), (xMax,yMax)) $ concat xs
  where
    xMax = length xs
    yMax = length $ head xs

heightmapP :: Parser Heightmap
heightmapP = mkHeightmap <$> row `endBy` newline
  where row = many $ digitToInt <$> digitChar

northOf (x,y) = (x-1,y)
eastOf (x,y) = (x, y+1)
southOf (x,y) = (x+1,y)
westOf (x,y) = (x, y-1)
directions = [northOf, eastOf, southOf, westOf]

adjPoints :: Heightmap -> Point -> [Point]
adjPoints hs = filter (inRange $ bounds hs) . sequenceA directions

getAdj :: Heightmap -> Point -> [Int]
getAdj hs = map (hs!) . adjPoints hs

lowPoints :: Heightmap -> [Point]
lowPoints hs = filter isLow $ indices hs
  where isLow p = all (> hs!p) $ getAdj hs p

riskLevel :: Heightmap -> Int
riskLevel hs = sum $ map ((+1) . (hs!)) $ lowPoints hs

basin :: Heightmap -> Point -> PointSet
basin hs p = basin' hs p S.empty
  where
    bs = bounds hs
    basin' hs p seen
      | not $ inRange bs p = seen
      | hs!p == 9          = seen
      | p `S.member` seen  = seen
      | otherwise          = foldr (basin' hs) (S.insert p seen) $ sequenceA directions p

largestBasins :: Heightmap -> [Int]
largestBasins hs = take 3 $ reverse $ sort $ map (S.size . basin hs) $ lowPoints hs

main = do
  runTests

  Right heightmap <- parse heightmapP "(stdin)" <$> getContents

  putStr "--\nPart 1: "
  print $ riskLevel heightmap
  putStr "Part 2: "
  print $ product $ largestBasins heightmap

runTests = hspec $ do
  it "gets adjacent elements" $ do
    sort (getAdj testSample (1, 1)) `shouldBe` [1, 3]
    sort (getAdj testSample (5, 3)) `shouldBe` [6, 8, 9]
    sort (getAdj testSample (3, 4)) `shouldBe` [5, 7, 7, 7]
  it "finds low points" $ do
    lowPoints testSample `shouldBe` [(1,2),(1,10),(3,3),(5,7)]
  it "calculates risk level" $ do
    riskLevel testSample `shouldBe` 15
  it "finds basins" $ do
    S.size (basin testSample (1,1)) `shouldBe` 3
    S.size (basin testSample (1,10)) `shouldBe` 9
    S.size (basin testSample (3,3)) `shouldBe` 14
    S.size (basin testSample (5,7)) `shouldBe` 9
  it "finds all basins" $ do
    product (largestBasins testSample) `shouldBe` 1134
testInput = "2199943210\n3987894921\n9856789892\n8767896789\n9899965678\n"
testSample = mkHeightmap [[2,1,9,9,9,4,3,2,1,0],
                          [3,9,8,7,8,9,4,9,2,1],
                          [9,8,5,6,7,8,9,8,9,2],
                          [8,7,6,7,8,9,6,7,8,9],
                          [9,8,9,9,9,6,5,6,7,8]]
