module Main where

import Test.Hspec
import Data.Either
import Data.List

brPair :: Char -> Char
brPair c = case c of '(' -> ')'; '[' -> ']'; '{' -> '}'; '<' -> '>'

complete :: String -> Either Char String
complete s = complete' s ""
  where
    complete' :: String -> String -> Either Char String
    complete' "" stack              = Right stack
    complete' (c:cs) stack
      | c `elem` ("([{<" :: String) = complete' cs (brPair c:stack)
      | c == head stack             = complete' cs $ tail stack
      | otherwise                   = Left c

scoreCorr :: Either Char a -> Int
scoreCorr (Left c) = case c of ')' -> 3; ']' -> 57; '}' -> 1197; '>' -> 25137
scoreCorr _ = 0

scoreAllCorr :: [String] -> Int
scoreAllCorr = sum . map (scoreCorr . complete)

scoreCompletion :: String -> Int
scoreCompletion = foldl combine 0
  where
    combine s c = s * 5 + score c
    score c = case c of
      ')' -> 1
      ']' -> 2
      '}' -> 3
      '>' -> 4

winningScore :: [String] -> Int
winningScore = middle . sort . map scoreCompletion . rights . map complete
  where middle xs = xs !! (length xs `div` 2)

------------------------------------------------------------------------------

main = do
  runTests

  input <- lines <$> getContents

  putStr "--\nPart 1: "
  print $ scoreAllCorr input
  putStr "Part 2: "
  print $ winningScore input

------------------------------------------------------------------------------

runTests = hspec $ do
  it "validates legal chunks" $ do
    complete "([])" `shouldSatisfy` isRight
    complete "{()()()}" `shouldSatisfy` isRight
    complete "<([{}])>" `shouldSatisfy` isRight
    complete "[<>({}){}[([])<>]]" `shouldSatisfy` isRight
    complete "(((((((((())))))))))" `shouldSatisfy` isRight
  it "catches corrupted chunks" $ do
    complete "(]" `shouldBe` Left ']'
    complete "{()()()>" `shouldBe` Left '>'
    complete "(((()))}" `shouldBe` Left '}'
    complete "<([]){()}[{}])" `shouldBe` Left ')'
  it "scores errors correctly" $ do
    scoreAllCorr testSample `shouldBe` 26397
  it "completes correctly" $ do
    complete "[({(<(())[]>[[{[]{<()<>>" `shouldBe` Right "}}]])})]"
    complete "[(()[<>])]({[<{<<[]>>(" `shouldBe` Right ")}>]})"
    complete "(((({<>}<{<{<>}{[]{[]{}" `shouldBe` Right "}}>}>))))"
    complete "{<[[]]>}<{[{[{[]{()[[[]" `shouldBe` Right "]]}}]}]}>"
    complete "<{([{{}}[<[[[<>{}]]]>[]]" `shouldBe` Right "])}>"
  it "scores completions correctly" $ do
    scoreCompletion "])}>" `shouldBe` 294
    scoreCompletion "}}]])})]" `shouldBe` 288957
    scoreCompletion ")}>]})" `shouldBe` 5566
    scoreCompletion "}}>}>))))" `shouldBe` 1480781
    scoreCompletion "]]}}]}]}>" `shouldBe` 995444
    scoreCompletion "])}>" `shouldBe` 294
  it "finds winning score" $ do
    winningScore testSample `shouldBe` 288957
testSample = ["[({(<(())[]>[[{[]{<()<>>",
              "[(()[<>])]({[<{<<[]>>(",
              "{([(<{}[<>[]}>{[]{[(<()>",
              "(((({<>}<{<{<>}{[]{[]{}",
              "[[<[([]))<([[{}[[()]]]",
              "[{[{({}]{}}([{[{{{}}([]",
              "{<[[]]>}<{[{[{[]{()[[[]",
              "[<(<(<(<{}))><([]([]()",
              "<{([([[(<>()){}]>(<<{{",
              "<{([{{}}[<[[[<>{}]]]>[]]"]
