module Main where

import Test.Hspec
import Prelude as P
import Data.Massiv.Array as A
import Data.List(find)
import Data.Maybe(fromJust)
import Data.Char(digitToInt)

type Grid = Array P Ix2 Int

parse :: String -> Grid
parse = fromLists' Seq . P.map (P.map digitToInt) . lines

inc :: Grid -> Grid
inc = compute . A.map (+1)

flashUpdate :: Stencil Ix2 Int Int
flashUpdate = makeStencil (Sz (3 :. 3)) (1 :. 1) go
  where
    go get
      | get (0 :. 0) == 0 = 0
      | get (0 :. 0) > 9  = 0
      | otherwise = get (0 :. 0)
                + P.sum (P.map (flashInc . get) [-1 :. -1, -1 :. 0, -1 :. 1, 0 :. -1, 0 :. 1, 1 :. -1, 1 :. 0, 1 :. 1])
    flashInc i = if i > 9 then 1 else 0

flash :: Grid -> Grid
flash g
  | A.all (<=9) g = g
  | otherwise     = flash $ update g
  where update = computeAs P . mapStencil (Fill 0) flashUpdate

step :: Grid -> Grid
step = flash . inc

flashes :: Grid -> Int
flashes = A.sum . A.map (fromEnum . (==0))

flashesAfter :: Int -> Grid -> Int
flashesAfter n = P.sum . P.take (n+1) . P.map flashes . iterate step

simultaneous :: Grid -> Int
simultaneous = fst . fromJust . find isSimul . P.zip [0..] . iterate step
  where isSimul (_, g) = flashes g == elemsCount g

------------------------------------------------------------------------------

main = do
  runTests

  grid <- parse <$> getContents

  putStr "--\nPart 1: "
  print $ flashesAfter 100 grid
  putStr "Part 2: "
  print $ simultaneous grid

------------------------------------------------------------------------------

runTests = hspec $ do
  it "parses input" $ do
    parse testInput `shouldBe` testSample
  it "steps simulation" $ do
    step testSample `shouldBe` testSample1
    step testSample1 `shouldBe` testSample2
  it "counts flashes" $ do
    flashes testSample `shouldBe` 0
    flashes testSample1 `shouldBe` 0
    flashes testSample2 `shouldBe` 35
  it "simulates and counts flashes" $ do
    flashesAfter 10 testSample `shouldBe` 204
    flashesAfter 100 testSample `shouldBe` 1656
  it "finds simultaneous flash" $ do
    simultaneous testSample `shouldBe` 195
testInput = "5483143223\n2745854711\n5264556173\n6141336146\n6357385478\n4167524645\n2176841721\n6882881134\n4846848554\n5283751526"
testSample = fromLists' Seq [[5,4,8,3,1,4,3,2,2,3]
                            ,[2,7,4,5,8,5,4,7,1,1]
                            ,[5,2,6,4,5,5,6,1,7,3]
                            ,[6,1,4,1,3,3,6,1,4,6]
                            ,[6,3,5,7,3,8,5,4,7,8]
                            ,[4,1,6,7,5,2,4,6,4,5]
                            ,[2,1,7,6,8,4,1,7,2,1]
                            ,[6,8,8,2,8,8,1,1,3,4]
                            ,[4,8,4,6,8,4,8,5,5,4]
                            ,[5,2,8,3,7,5,1,5,2,6]]
testSample1 = parse "6594254334\n3856965822\n6375667284\n7252447257\n7468496589\n5278635756\n3287952832\n7993992245\n5957959665\n6394862637"
testSample2 = parse "8807476555\n5089087054\n8597889608\n8485769600\n8700908800\n6600088989\n6800005943\n0000007456\n9000000876\n8700006848"
