module Main where

import Test.Hspec
import Control.Monad(guard)
import Data.Char(isAsciiUpper)
import Data.Either(isRight,fromRight)
import Data.List(nub)
import Data.Map.Strict(Map,fromListWith,(!?))
import qualified Data.Map.Strict as M
import Data.Void(Void)
import Text.Megaparsec
import Text.Megaparsec.Char

data Cave = Big String | Small String | Start | End
  deriving (Eq, Ord)
type Path = [Cave]
type Links = Map Cave [Cave]
type Parser = Parsec Void String

instance Show Cave where
  show Start = "start"
  show End = "end"
  show (Big name) = name
  show (Small name) = name

isBig :: Cave -> Bool
isBig (Big _) = True
isBig _ = False

isSmall :: Cave -> Bool
isSmall (Small _) = True
isSmall _ = False

isEnd :: Cave -> Bool
isEnd End = True
isEnd _ = False

noSmolDups :: Path -> Bool
noSmolDups path = length smalls == length (nub smalls)
  where smalls = filter isSmall path

mkCave :: String -> Cave
mkCave "start" = Start
mkCave "end" = End
mkCave s
  | all isAsciiUpper s = Big s
  | otherwise          = Small s

linksP :: Parser Links
linksP = fromListWith (++) . concat <$> link `sepEndBy` newline
  where
    link = bidirectional <$> cave <* char '-' <*> cave
    cave = mkCave <$> many letterChar
    bidirectional a b = [(a, [b]), (b, [a])]

extend :: Links -> (Cave -> Path -> Bool) -> [Path] -> [Path]
extend links canGo paths = do
  path@(current:_) <- paths
  guard $ not $ isEnd current

  let Just options = links !? current
  next <- options
  guard $ canGo next path

  return $ next:path

isComplete :: Path -> Bool
isComplete [] = False
isComplete ps = head ps == End && last ps == Start

findPaths :: Links -> (Cave -> Path -> Bool) -> [Path]
findPaths links canGo = filter isComplete $ concat $ takeWhile (not . null)
  $ iterate (extend links canGo) [[Start]]

findPaths1 :: Links -> [Path]
findPaths1 links = findPaths links canGo
  where canGo next history = isBig next || next `notElem` history

findPaths2 :: Links -> [Path]
findPaths2 links = findPaths links canGo
  where
    canGo next history =
      let visits = length $ filter (==next) history in
        isBig next || isEnd next
        || (isSmall next && if noSmolDups history then visits < 2 else visits == 0)

------------------------------------------------------------------------------

main = do
  runTests

  Right input <- parse linksP "(stdin)" <$> getContents

  putStr "--\nPart 1: "
  print $ length $ findPaths1 input
  putStr "Part 2: "
  print $ length $ findPaths2 input

------------------------------------------------------------------------------

runTests = hspec $ do
  it "parses input" $ do
    parse linksP "(test)" testInput `shouldBe` Right testSample
    parse linksP "(test)" testInput1 `shouldSatisfy` isRight
    parse linksP "(test)" testInput2 `shouldSatisfy` isRight
  it "finds valid paths" $ do
    length (findPaths1 testSample) `shouldBe` 10
    length (findPaths1 testSample1) `shouldBe` 19
    length (findPaths1 testSample2) `shouldBe` 226
  it "finds other valid paths" $ do
    -- length (findPaths2 testSample) `shouldBe` 36
    length (findPaths2 testSample1) `shouldBe` 103
    -- length (findPaths2 testSample2) `shouldBe` 3509

testInput = "start-A\nstart-b\nA-c\nA-b\nb-d\nA-end\nb-end"
testSample = M.fromList [(Big "A",   [End,Small "b", Small "c", Start])
                        ,(Small "b", [End, Small "d", Big "A", Start])
                        ,(Small "c", [Big "A"])
                        ,(Small "d", [Small "b"])
                        ,(End,       [Small "b", Big "A"])
                        ,(Start,     [Small "b", Big "A"])]
testInput1 = "dc-end\nHN-start\nstart-kj\ndc-start\ndc-HN\nLN-dc\nHN-end\nkj-sa\nkj-HN\nkj-dc"
testSample1 = fromRight M.empty $ parse linksP "(test)" testInput1
testInput2 = "fs-end\nhe-DX\nfs-he\nstart-DX\npj-DX\nend-zg\nzg-sl\nzg-pj\npj-he\nRW-he\nfs-DX\npj-RW\nzg-RW\nstart-pj\nhe-WI\nzg-he\npj-fs\nstart-RW"
testSample2 = fromRight M.empty $ parse linksP "(test)" testInput2
